@@ -30,28 +30,30 @@
    and the righthandside values of the constraints the linked list is in
    reversed order with respect to the rownumbers
    ============================================================================ */
-#include "lpkit.h"
-#include "lpglob.h"
 #include <string.h>
 #include <limits.h>
+#include "lpkit.h"
+#include "lpglob.h"
 
 short            *relat;
 int              Verbose;
 constraint_name  *First_constraint_name;
-rside            *First_rside;
+rside            *First_rside, *rs;
 tmp_store_struct tmp_store;
 short            Ignore_decl;
-hashtable        *Hash_tab;
-
-#define HASHSIZE  10007 /* prime number is better, MB */
+hashstruct       *Hash_tab;
 
 /*
- * errorhandeling routine for yyparse()
+ * error handling routine for yyparse()
  */
 void yyerror(char *string)
 {
-  fprintf(stderr, "PARSING ERROR: %s on line %d, quiting\n", string, yylineno);
-  exit(EXIT_FAILURE);
+  report(NULL, CRITICALSTOP, string);
+}
+
+int yywrap() /* supply a default yywrap() function */
+{
+  return(1);
 }
 
 void check_decl(int within_int_decl)
@@ -60,8 +62,9 @@ void check_decl(int within_int_decl)
     Ignore_decl = FALSE;
   }
   else {
-    fprintf(stderr, "Unknown declaration specifier on line %d, ignored\n",
-	    yylineno);
+    if (Verbose >= NORMAL)
+      report(NULL, NORMAL, "Unknown declaration specifier on line %d, ignored",
+	      yylineno);
     Ignore_decl = TRUE;
   }
 }
@@ -69,33 +72,45 @@ void check_decl(int within_int_decl)
 void add_int_var(char *name)
 {
   hashelem *hp;
-  
-  if(Verbose)
-    fprintf(stderr, "int: %s\n", name);
-  if(!(hp = findhash(name, Hash_tab)))
-    fprintf(stderr,
-	    "Unknown variable %s declared integer on line %d, ignored\n",
-	    name, yylineno);
-  else if(hp->must_be_int)
-    fprintf(stderr, "Variable %s declared integer more than once on line %d\n",
-	    name, yylineno);
+
+  if((hp = findhash(name, Hash_tab)) == NULL) {
+    if (Verbose >= NORMAL)
+      report(NULL, NORMAL,
+ 	      "Unknown variable %s declared integer on line %d, ignored",
+	      name, yylineno);
+  }
+  else if(hp->must_be_int) {
+    if (Verbose >= NORMAL)
+      report(NULL, NORMAL, "Variable %s declared integer more than once on line %d",
+	      name, yylineno);
+  }
   else
     hp->must_be_int = TRUE;
 }
 
 /*
- * initialisation of hashtable and globals.
+ * initialisation of hashstruct and globals.
  */
-void init_read(void)
+int init_read(void)
 {
+  int ok = FALSE;
+
   Rows = 0;
   Non_zeros = 0;
   Columns = 0;
-  CALLOC(First_rside, 1);
-  First_rside->value = 0;
-  /* first row (nr 0) is always the objective function */
-  First_rside->relat = OF;
-  Hash_tab = create_hash_table(HASHSIZE);
+  if (CALLOC(First_rside, 1) != NULL) {
+    rs = First_rside;
+    rs->value = rs->range_value = 0;
+    /* first row (nr 0) is always the objective function */
+    rs->relat = OF;
+    rs->range_relat = -1;
+    if ((Hash_tab = create_hash_table(HASHSIZE)) == NULL) {
+      FREE(First_rside);
+    }
+    else
+      ok = TRUE;
+  }
+  return(ok);
 } /* init */
 
 
@@ -122,9 +137,11 @@ static column *getrow(column *p,
 static bound *create_bound_rec(void)
 {
   bound *bp;
-  CALLOC(bp, 1);
-  bp->upbo = DEF_INFINITE;
-  bp->lowbo = 0;
+
+  if (CALLOC(bp, 1) != NULL) {
+    bp->upbo = DEF_INFINITE;
+    bp->lowbo = 0;
+  }
   return(bp);
 } /* create_bound_rec */
 
@@ -147,30 +164,34 @@ void null_tmp_store(void)
  *	      A(row, variable). If A(row, variable) already contains data,
  *	      value is added to the existing value.
  */
-static void store(char *variable,
+static int store(char *variable,
 		  int row,
-		  REAL value) 
-{ 
+		  REAL value)
+{
   hashelem *h_tab_p;
   column *col_p;
-  
+
   if(value == 0) {
-    fprintf(stderr,
-	    "(store) Warning, variable %s has an effective coefficient of 0 on line %d. Ignored.\n",
-	    variable, yylineno);
-    return;
+    if (Verbose >= NORMAL)
+      report(NULL, NORMAL,
+	      "(store) Warning, variable %s has an effective coefficient of 0 on line %d. Ignored.",
+	      variable, yylineno);
+    return(TRUE);
   }
-  
+
   if((h_tab_p = findhash(variable, Hash_tab)) == NULL) {
-    h_tab_p = puthash(variable, Hash_tab);
+    if (((h_tab_p = puthash(variable, Hash_tab)) == NULL) ||
+        (CALLOC(h_tab_p->col, 1) == NULL)
+       ) return(FALSE);
     Columns++; /* counter for calloc of final array */
-    CALLOC(h_tab_p->col, 1);
+
     Non_zeros++; /* for calloc of final arrays */
     h_tab_p->col->row = row;
     h_tab_p->col->value = value;
   }
   else if((col_p = getrow(h_tab_p->col, row)) == NULL) {
-    CALLOC(col_p, 1);
+    if (CALLOC(col_p, 1) == NULL)
+      return(FALSE);
     Non_zeros++; /* for calloc of final arrays */
     col_p->value = value;
     col_p->row = row;
@@ -179,6 +200,7 @@ static void store(char *variable,
   }
   else
     col_p->value += value;
+  return(TRUE);
 } /* store */
 
 /*
@@ -191,28 +213,48 @@ void store_re_op(void)
   short tmp_relat;
 
   switch(yytext[0]) {
-    
+
   case '=':
     tmp_relat = EQ;
     break;
-    
+
   case '>':
     tmp_relat=GE;
     break;
-    
+
   case '<':
     tmp_relat=LE;
     break;
-    
+
   default:
-    fprintf(stderr, "Error: unknown relational operator %s on line %d\n",
-	    yytext, yylineno);
-    exit(EXIT_FAILURE);
+    report(NULL, CRITICALSTOP, "Error: unknown relational operator %s on line %d",
+	   yytext, yylineno);
+    return;
     break;
   }
 
   if(Lin_term_count > 1) /* it is not a bound */
-    First_rside->relat = tmp_relat;
+    rs->relat = tmp_relat;
+  else if(Lin_term_count == 0) { /* it is a range */
+    if(rs == NULL) { /* range before row, already reported */
+      if(Verbose >= CRITICAL)
+        report(NULL, CRITICAL, "Error on line %d: range for undefined row.",
+  	        yylineno);
+    }
+    else if(rs->range_relat != -1) {
+      if(Verbose >= CRITICAL)
+        report(NULL, CRITICAL, "Error on line %d: There was already a range for this row.",
+  	        yylineno);
+    }
+    else if(tmp_relat == rs->relat) {
+      if(Verbose >= CRITICAL)
+        report(NULL, CRITICAL, "Error on line %d: relational operator for range is the same as relation operator for equation.",
+  	        yylineno);
+      rs->range_relat = -2;
+    }
+    else
+      rs->range_relat = tmp_relat;
+  }
   else /* could be a bound */
     tmp_store.relat = tmp_relat;
 } /* save_re_op */
@@ -224,7 +266,23 @@ void store_re_op(void)
 void rhs_store(REAL value)
 {
   if(Lin_term_count > 1) /* not a bound */
-    First_rside->value += value;
+    rs->value += value;
+  else if(Lin_term_count == 0) { /* a range */
+    if(rs == NULL); /* range before row, already reported */
+    else if(rs->range_relat < 0) /* was a bad range; ignore */;
+    else if(((rs->relat == LE) && (rs->range_relat == GE) &&
+         (rs->value < value)) ||
+        ((rs->relat == GE) && (rs->range_relat == LE) &&
+         (rs->value > value)) ||
+	((rs->relat == EQ) || (rs->range_relat == EQ))) {
+      if(Verbose >= CRITICAL)
+        report(NULL, CRITICAL, "Error on line %d: range restriction is conflicting with equation on previous line",
+  	        yylineno);
+      rs->range_relat = -2;
+    }
+    else
+      rs->range_value += value;
+  }
   else /* a bound */
     tmp_store.rhs_value += value;
 } /* RHS_store */
@@ -234,54 +292,59 @@ void rhs_store(REAL value)
  * count the amount of lineair terms in a constraint
  * only store in data-structure if the constraint is not a bound
  */
-void var_store(char *var, int row, REAL value)
+int var_store(char *var, int row, REAL value)
 {
   if(strlen(var) > MAXSTRL) {
-    fprintf(stderr,
-	    "Variable name '%s' too long, at most %d characters allowed\n",
-	    var, MAXSTRL);
-    exit(EXIT_FAILURE);
+    if (Verbose >= CRITICAL)
+      report(NULL, CRITICAL, "Variable name '%s' too long, at most %d characters allowed",
+	      var, MAXSTRL);
+    return(FALSE);
   }
   /* also in a bound the same var name can occur more than once. Check for
      this. Don't increment Lin_term_count */
-  
+
   if(Lin_term_count != 1 || strcmp(tmp_store.name, var) != 0)
     Lin_term_count++;
 
   /* always store objective function with rownr == 0. */
-  if(row == 0) {
-    store(var,  row,  value);
-    return;
-  }
-  
+  if(row == 0)
+    return(store(var,  row,  value));
+
   if(Lin_term_count == 1) { /* don't store yet. could be a bound */
     strcpy(tmp_store.name, var);
     tmp_store.row = row;
     tmp_store.value += value;
-    return;
+    return(TRUE);
   }
-  
+
   if(Lin_term_count == 2) { /* now you can also store the first variable */
     rside *rp;
-    
+
     /* make space for the rhs information */
-    CALLOC(rp, 1);
+    if (CALLOC(rp, 1) == NULL)
+      return(FALSE);
     rp->next = First_rside;
-    First_rside = rp;
-    First_rside->value = tmp_store.rhs_value;
-    First_rside->relat = tmp_store.relat;
-    
-    if(tmp_store.value != 0)
-      store(tmp_store.name, tmp_store.row, tmp_store.value);
-    else
-      fprintf(stderr,
-	      "Warning, variable %s has an effective coefficient of 0 on line %d. Ignored.\n",
-	      tmp_store.name, yylineno);
-    
+    First_rside = rs = rp;
+    rs->row = row;
+    rs->value = tmp_store.rhs_value;
+    rs->relat = tmp_store.relat;
+    rs->range_relat = -1;
+
+    if(tmp_store.value != 0) {
+      if (!store(tmp_store.name, tmp_store.row, tmp_store.value))
+        return(FALSE);
+    }
+    else {
+      if (Verbose >= NORMAL)
+        report(NULL, NORMAL,
+	        "Warning, variable %s has an effective coefficient of 0 on line %d. Ignored.",
+	        tmp_store.name, yylineno);
+    }
+
     null_tmp_store();
   }
-  
-  store(var, row, value);
+
+  return(store(var, row, value));
 } /* var_store */
 
 
@@ -289,25 +352,28 @@ void var_store(char *var, int row, REAL value)
 /*
  * store the information in tmp_store because it is a bound
  */
-void store_bounds(void)
+int store_bounds(void)
 {
   if(tmp_store.value != 0) {
     hashelem *h_tab_p;
     REAL boundvalue;
-      
+
     if((h_tab_p = findhash(tmp_store.name, Hash_tab)) == NULL) {
       /* a new columnname is found, create an entry in the hashlist */
-      h_tab_p = puthash(tmp_store.name, Hash_tab);
+      if ((h_tab_p = puthash(tmp_store.name, Hash_tab)) == NULL)
+        return(FALSE);
       Columns++; /* counter for calloc of final array */
       /* create a place to store bounds information */
-      h_tab_p->bnd = create_bound_rec();
+      if ((h_tab_p->bnd = create_bound_rec()) == NULL)
+        return(FALSE);
     }
     else if(h_tab_p->bnd == NULL)
       /* create a place to store bounds information */
-      h_tab_p->bnd = create_bound_rec();
+      if ((h_tab_p->bnd = create_bound_rec()) == NULL)
+        return(FALSE);
 
     /* else bound_rec already exists */
-      
+
     if(tmp_store.value < 0) { /* divide by negative number, */
       /* relational operator may change */
       if(tmp_store.relat == GE)
@@ -319,72 +385,98 @@ void store_bounds(void)
     boundvalue = tmp_store.rhs_value / tmp_store.value;
     if(   ((tmp_store.relat == EQ) && (boundvalue < 0))
        || ((tmp_store.relat == LE) && (boundvalue < 0))) { /* Error */
-      fprintf(stderr,
-	      "Error on line %d: variables must always be non-negative\n",
-	      yylineno);
-      exit(EXIT_FAILURE);
+      if (Verbose >= CRITICAL)
+        report(NULL, CRITICAL, "Error on line %d: variables must always be non-negative",
+	        yylineno);
+      return(FALSE);
     }
-      
+
     if((tmp_store.relat == GE) && (boundvalue <= 0)) /* Warning */
-      fprintf(stderr,
-	      "Warning on line %d: useless bound; variables are always >= 0\n",
-	      yylineno);
-      
+      if (Verbose >= NORMAL)
+        report(NULL, NORMAL,
+	        "Warning on line %d: useless bound; variables are always >= 0",
+	        yylineno);
+
     /* bound seems to be sane, add it */
     if((tmp_store.relat == GE) || (tmp_store.relat == EQ)) {
       if(h_tab_p->bnd->lowbo <  boundvalue)
 	h_tab_p->bnd->lowbo = boundvalue;
       else
-	fprintf(stderr, "Ineffective lower bound on line %d, ignored\n",
-		yylineno);
+        if (Verbose >= NORMAL)
+          report(NULL, NORMAL, "Ineffective lower bound on line %d, ignored",
+		  yylineno);
     }
     if((tmp_store.relat == LE) || (tmp_store.relat == EQ)) {
       if(h_tab_p->bnd->upbo >  boundvalue)
 	h_tab_p->bnd->upbo  = boundvalue;
       else
-	fprintf(stderr, "Ineffective upper bound on line %d, ignored\n",
-		yylineno);
+        if (Verbose >= NORMAL)
+          report(NULL, NORMAL, "Ineffective upper bound on line %d, ignored",
+		  yylineno);
     }
-      
+
     /* check for empty range */
     if(h_tab_p->bnd->upbo < h_tab_p->bnd->lowbo) {
-      fprintf(stderr,
-	      "Error: bound on line %d contradicts earlier bounds, exiting\n",
-	      yylineno);
-      exit(EXIT_FAILURE);
+      if (Verbose >= CRITICAL)
+        report(NULL, CRITICAL, "Error: bound on line %d contradicts earlier bounds, exiting",
+	        yylineno);
+      return(FALSE);
     }
   }
   else /* tmp_store.value = 0 ! */ {
-    fprintf(stderr,
-	    "Error, variable %s has an effective coefficient of 0 in bound on line %d. Exiting.\n",
-	    tmp_store.name, yylineno);
-    exit(EXIT_FAILURE);
+    if (Verbose >= CRITICAL)
+      report(NULL, CRITICAL, "Error, variable %s has an effective coefficient of 0 in bound on line %d. Exiting.",
+	      tmp_store.name, yylineno);
+    return(FALSE);
   }
-  
+
   null_tmp_store();
+  return(TRUE);
 } /* store_bounds */
 
-void add_constraint_name(char *name, int row)
+int add_constraint_name(char *name, int row)
 {
   constraint_name *cnp;
 
   if(!First_constraint_name) { /* first time only */
-    CALLOC(First_constraint_name, 1);
+    if (CALLOC(First_constraint_name, 1) == NULL)
+      return(FALSE);
+    cnp = First_constraint_name;
+    rs = NULL;
   }
   else {
     cnp = First_constraint_name;
-    CALLOC(First_constraint_name, 1);
-    First_constraint_name->next = cnp;
+    while(cnp) {
+      if (strcmp(cnp->name, name) == 0)
+        break;
+      cnp = cnp->next;
+    }
+    if (cnp) {
+      row = cnp->row;
+
+      rs = First_rside;
+      while ((rs != NULL) && (rs->row != row))
+        rs = rs->next;
+    }
+    else {
+      cnp = First_constraint_name;
+      if (CALLOC(First_constraint_name, 1) == NULL)
+        return(FALSE);
+      First_constraint_name->next = cnp;
+      cnp = First_constraint_name;
+      rs = NULL;
+    }
   }
-  strcpy(First_constraint_name->name, name);
-  First_constraint_name->row = row;
+  strcpy(cnp->name, name);
+  cnp->row = row;
+  return(TRUE);
 }
 
 /*
  * transport the data from the intermediate structure to the sparse matrix
  * and free the intermediate structure
  */
-void readinput(lprec *lp)
+int readinput(lprec *lp)
 {
   int    i, j, index, nn_ind;
   column *cp,*tcp; /* tcp (temporary cp) points to memory-space to free */
@@ -392,28 +484,49 @@ void readinput(lprec *lp)
   bound *bp;
   rside *rp;
   constraint_name *cnp;
-   
+  REAL *row;
+
+  if (MALLOC(row, 1 + Rows) == NULL)
+    return(FALSE);
+
   /* fill names with the rownames */
   for(cnp = First_constraint_name; cnp; cnp = cnp->next)
-    strcpy(lp->row_name[cnp->row], cnp->name);
+    if (!set_row_name(lp, cnp->row, cnp->name)) {
+      free(row);
+      return(FALSE);
+    }
 
   for(i = Rows;i >= 0;i--) {
     rp = First_rside;
+    if ((rp->range_relat >= 0) && (rp->range_value == rp->value)) {
+      rp->relat = EQ;
+      rp->range_relat = EQ;
+    }
     relat[i] = rp->relat;
     lp->orig_rh[i] = rp->value;
+    if (rp->range_relat >= 0)
+      lp->orig_upbo[i] = my_abs(rp->range_value - rp->value);
     First_rside = rp->next;
     free(rp); /* free memory when data has been read */
   }
-  
+
   /* change upperbound to zero if the relational operator is the equal sign */
   for(i = 1; i <= Rows; i++)
     if(relat[i] == EQ)
       lp->orig_upbo[i] = 0;
 
+/*
   for(i = 0; i <= Rows; i++)
-    if(strcmp(lp->row_name[i], "")==0)
-      sprintf(lp->row_name[i],"r_%d",i); 
-  
+    if((!lp->names_used) || (lp->row_name[i] == NULL) || (strcmp(lp->row_name[i], "")==0)) {
+      char holdstr[NAMELEN];
+      sprintf(holdstr,ROWNAMEMASK,i);
+      if (!set_row_name(lp,i,holdstr)) {
+        free(row);
+        return(FALSE);
+      }
+    }
+*/
+
   /* start reading the Hash_list structure */
   index = 0;
   nn_ind = 0;
@@ -423,10 +536,11 @@ void readinput(lprec *lp)
     while(hp != NULL) {
       /* put an index in the cend array when a new name is found */
       lp->col_end[index++] = nn_ind;
-      
+
       /* check if it must be an integer variable */
       if(hp->must_be_int) {
-	lp->must_be_int[Rows + index]=TRUE;
+	/* lp->must_be_int[Rows + index]=TRUE; */
+        set_int(lp, index, TRUE);
       }
       /* check for bound */
       if(hp->bnd != NULL) {
@@ -435,16 +549,21 @@ void readinput(lprec *lp)
 	lp->orig_upbo[Rows+index] = bp->upbo;
 	free(bp); /* free memory when data has been read*/
       }
-	  
+
       /* copy name of column variable */
-      strcpy(lp->col_name[index], hp->name);
-	  
-      /* put matrix values in sparse matrix */
+      /* lp->col_name[index] = hp->name; */
+      if (!set_col_name(lp, index, hp->name)) {
+	free(row);
+        return(FALSE);
+      }
+
+      for(j = 0; j <= Rows; j++)
+        row[j] = 0.0;
+
+      /* put matrix values in intermediate row */
       cp = hp->col;
       while(cp!=NULL) {
-	lp->mat[nn_ind].row_nr = cp->row;
-	lp->mat[nn_ind].value = cp->value;
-	nn_ind++;
+	row[cp->row] = cp->value;
 	tcp = cp;
 	cp = cp->next;
 	free(tcp); /* free memory when data has been read */
@@ -453,14 +572,26 @@ void readinput(lprec *lp)
       hp = hp->next;
       free(thp->name);
       free(thp); /* free memory when data has been read */
+
+      /* put matrix values in sparse matrix */
+      /* this makes sure that values are in order in the sparse matrix. This is a requirement */
+      for(j = 0; j <= Rows; j++)
+        if(row[j]) {
+	  lp->mat[nn_ind].row_nr = j;
+	  lp->mat[nn_ind].value = row[j];
+	  nn_ind++;
+        }
     }
-	Hash_tab->table[i] = NULL;
+    Hash_tab->table[i] = NULL;
   }
-  lp->col_end[index] = nn_ind; 
+  lp->col_end[index] = nn_ind;
 
   /* the following should be replaced by a call to the MPS print routine MB */
-  
+
+#if 0
   if(Verbose) {
+    int j;
+
     printf("\n");
     printf("**********Data read**********\n");
     printf("Rows    : %d\n", Rows);
@@ -477,181 +608,117 @@ void readinput(lprec *lp)
 	printf(" G  ");
       else if(relat[i] == OF)
 	printf(" N  ");
-      printf("%s\n", lp->row_name[i]);
+      printf("%s\n", get_row_name(lp, i));
     }
-    
+
     printf("COLUMNS\n");
     j = 0;
     for(i = 0; i < Non_zeros; i++) {
       if(i == lp->col_end[j])
 	j++;
-      printf("    %-8s  %-8s  %g\n", lp->col_name[j],
-	     lp->row_name[lp->mat[i].row_nr], (double)lp->mat[i].value);
+      printf("    %-8s  %-8s  %g\n", get_col_name(lp, j),
+	     get_row_name(lp, lp->mat[i].row_nr), (double)lp->mat[i].value);
     }
 
     printf("RHS\n");
     for(i = 0; i <= Rows; i++) {
-      printf("    RHS       %-8s  %g\n", lp->row_name[i],
+      printf("    RHS       %-8s  %g\n", get_row_name(lp, i),
 	     (double)lp->orig_rh[i]);
     }
-      
+
     printf("RANGES\n");
     for(i = 1; i <= Rows; i++)
       if((lp->orig_upbo[i] != lp->infinite) && (lp->orig_upbo[i] != 0)) {
-	printf("    RGS       %-8s  %g\n", lp->row_name[i],
+	printf("    RGS       %-8s  %g\n", get_row_name(lp, i),
 	       (double)lp->orig_upbo[i]);
       }
       else if((lp->orig_lowbo[i] != 0)) {
-	printf("    RGS       %-8s  %g\n", lp->row_name[i],
+	printf("    RGS       %-8s  %g\n", get_row_name(lp, i),
 	       (double)-lp->orig_lowbo[i]);
       }
- 
+
     printf("BOUNDS\n");
-    for(i = Rows + 1; i <= Sum; i++) {
-      if(lp->orig_upbo[i] < lp->infinite)
-	printf(" UP BND       %-8s  %g\n", lp->col_name[i - Rows],
-	       (double)lp->orig_upbo[i]);
-      if(lp->orig_lowbo[i] > 0)
-	printf(" LO BND       %-8s  %g\n", lp->col_name[i - Rows],
-	       (double)lp->orig_lowbo[i]);
+    for(i = Rows + 1; i <= Rows + Columns; i++) {
+      if((lp->orig_lowbo[i] != 0) && (lp->orig_upbo[i] < lp->infinite) &&
+         (lp->orig_lowbo[i] == lp->orig_upbo[i])) {
+        printf(" FX BND       %-8s  %g\n", get_col_name(lp, i - Rows),
+               (double)lp->orig_upbo[i]);
+      }
+      else {
+        if(lp->orig_upbo[i] < lp->infinite)
+  	  printf(" UP BND       %-8s  %g\n", get_col_name(lp, i - Rows),
+  		 (double)lp->orig_upbo[i]);
+        if(lp->orig_lowbo[i] > 0)
+  	  printf(" LO BND       %-8s  %g\n", get_col_name(lp, i - Rows),
+  		 (double)lp->orig_lowbo[i]);
+      }
     }
-    
+
     printf("ENDATA\n");
   }
+#endif
+
+  free(row);
+  return(TRUE);
 } /* readinput */
 
-lprec *read_lp_file(FILE *input, short verbose, nstring lp_name)
+lprec *read_LP(char *input, short verbose, char *lp_name)
 {
-  lprec *lp;
+  FILE *fpin;
+  lprec *lp = NULL;
+
+  if((fpin = fopen(input, "r")) != NULL) {
+    lp = read_lp_file(fpin, verbose, lp_name);
+    fclose(fpin);
+  }
+  return(lp);
+}
+
+lprec *read_lp_file(FILE *input, short verbose, char *lp_name)
+{
+  lprec *lp = NULL;
   int i;
-  Verbose = verbose; 
+
+  Verbose = verbose;
 
   yyin = input;
-  Maximise = TRUE;  
+  Maximise = TRUE;
   yyparse();
-  fclose(input);
-  
-  CALLOC(lp, 1);
 
   Rows--;
-  Sum = Rows + Columns;
 
-  strcpy(lp->lp_name, lp_name);
+  if (CALLOC(relat, Rows + 1) != NULL) {
+    lp = make_lpext(Rows, Columns, Non_zeros, Non_zeros, lp_name);
+    if (lp != NULL) {
+      lp->verbose = verbose;
 
-  lp->verbose         = FALSE;
-  lp->print_duals     = FALSE;
-  lp->print_sol       = FALSE;
-  lp->debug           = FALSE;
-  lp->print_at_invert = FALSE;
-  lp->trace           = FALSE;
+      if (!readinput(lp)) {
+        delete_lp(lp);
+	lp = NULL;
+      }
+      else  {
+	if(Maximise)
+	  set_maxim(lp);
 
-  lp->rows          = Rows;
-  lp->columns       = Columns;
-  lp->sum           = Sum;
-  lp->rows_alloc    = Rows;
-  lp->columns_alloc = Columns;
-  lp->sum_alloc     = lp->sum;
+	for(i = 1; i <= Rows; i++) {
+	  REAL a = lp->orig_upbo[i]; /* keep upper bound (range) */
+	  set_constr_type(lp, i, relat[i]);
+	  lp->orig_upbo[i] = a;      /* restore upper bound (range) */
+	}
 
-  CALLOC(lp->row_name, lp->rows + 1);
-  CALLOC(lp->col_name, lp->columns + 1);
-  
-  lp->names_used    = TRUE;
-  lp->obj_bound     = DEF_INFINITE;
-  lp->bb_rule       = FIRST_NI;
-  lp->break_at_int  = FALSE;
-  lp->infinite      = DEF_INFINITE;
-  lp->epsilon       = DEF_EPSILON;
-  lp->epsb          = DEF_EPSB;
-  lp->epsd          = DEF_EPSD;
-  lp->epsel         = DEF_EPSEL;
-  lp->non_zeros     = Non_zeros;
-  lp->mat_alloc     = Non_zeros;
-  lp->row_end_valid = FALSE;
-  
-  MALLOC(lp->mat,         Non_zeros);
-  CALLOC(lp->col_no,      Non_zeros + 1);
-  CALLOC(lp->col_end,     Columns + 1);
-  CALLOC(lp->row_end,     Rows + 1);
-  CALLOC(lp->orig_rh,     Rows + 1);
-  CALLOC(lp->rh,          Rows + 1);
-  CALLOC(lp->rhs,         Rows + 1);
-  CALLOC(lp->must_be_int, Sum + 1);
-  MALLOC(lp->orig_upbo,   Sum + 1);
-  CALLOC(lp->upbo,        Sum + 1);
-  CALLOC(lp->orig_lowbo,  Sum + 1);
-  CALLOC(lp->lowbo,       Sum + 1);
+	/* lets free the temporary list of constraint names */
+	while(First_constraint_name) {
+	  constraint_name *cp;
 
-  for(i = 0; i <= Sum; i++) {
-    lp->orig_upbo[i]  = lp->infinite;
-    lp->orig_lowbo[i] = 0;
+	  cp = First_constraint_name;
+	  First_constraint_name = First_constraint_name->next;
+	  free(cp);
+	}
+
+	free_hash_table(Hash_tab);
+      }
+    }
+    free(relat);
   }
-
-  lp->basis_valid = TRUE;
-
-  CALLOC(lp->bas,   Rows + 1);
-  CALLOC(lp->basis, Sum + 1);
-  CALLOC(lp->lower, Sum + 1);
-
-  for(i = 0; i <= Rows; i++) {
-    lp->bas[i]   = i;
-    lp->basis[i] = TRUE;
-  }
-
-  for(i = Rows + 1; i <= Sum; i++)
-    lp->basis[i] = FALSE;
-
-  for(i = 0 ; i <= Sum; i++)
-    lp->lower[i] = TRUE;
- 
-  lp->eta_valid   = TRUE;
-  lp->eta_size    = 0;
-  lp->eta_alloc   = INITIAL_MAT_SIZE;
-  lp->max_num_inv = DEFNUMINV;
-
-  CALLOC(lp->eta_value,   INITIAL_MAT_SIZE + 1);
-  CALLOC(lp->eta_row_nr,  INITIAL_MAT_SIZE + 1);
-  CALLOC(lp->eta_col_end, Rows + lp->max_num_inv + 1);
-
-  lp->iter       = 0;
-  lp->total_iter = 0;
-
-  CALLOC(lp->solution,      Sum + 1);
-  CALLOC(lp->best_solution, Sum + 1);
-  CALLOC(lp->duals,         Rows + 1);
-
-  lp->maximise     = FALSE;
-  lp->floor_first  = TRUE;
-  lp->scaling_used = FALSE;
-
-  CALLOC(lp->ch_sign, Rows + 1);
-
-  for(i = 0; i <= Rows; i++)
-    lp->ch_sign[i] = FALSE;
-
-  lp->valid = FALSE; 
-
-  CALLOC(relat, Rows + 1);
-
-  readinput(lp);
-
-  if(Maximise)
-    set_maxim(lp);
-
-  for(i = 1; i <= Rows; i++)
-    set_constr_type(lp, i, relat[i]);
-
-  free(relat);
-
-  /* lets free the temporary list of constraint names */
-  while(First_constraint_name) {
-    constraint_name *cp;
-    
-    cp = First_constraint_name;
-    First_constraint_name = First_constraint_name->next;
-    free(cp);
-  }
-
-  free_hash_table(Hash_tab);
-
-  return(lp); 
+  return(lp);
 }
